
/*
 * This is a demo platform module.
 *
 * Every platform is a ES2015 module which must expose:
 * - Account (a subclass of 'Account'; a bit confusing, I know)
 * - name (a String)
 * - defaultBadge (a short String)
 * - signUp (a function that must return a Promise to a set of config options
 *      for a new configured account)
 */


import Account from '../abstracts/account';

import inquirer from 'inquirer';


// 🍂class ClockAccount
// 🍂inherits Account
// A dumb platform which displays timestamps.

class ClockAccount extends Account {

	constructor(callback, vorpal, opts) {
		this._timers = {};

		this._opts = Object.assign({
			colour: 'red'
		}, opts);

		return super.constructor(callback, opts);
	}

	get defaults() {
		return this._defaults;
	}


	// 🍂channel day; Ticks once every day
	// 🍂channel hour; Ticks once every hour
	// 🍂channel quarter; Ticks once every 15 minutes
	// 🍂channel minute; Ticks once every minute
	// 🍂channel second; Ticks once every second
	listen(channel) {

		let delay = Number(channel) * 1000;

		if (channel === 'day')     { delay = 1000 * 3600 * 24; }
		if (channel === 'hour')    { delay = 1000 * 3600; }
		if (channel === 'quarter') { delay = 1000 * 60 * 15; }
		if (channel === 'minute')  { delay = 1000 * 60; }
		if (channel === 'second')  { delay = 1000; }

		if (delay <= 0) { return; }

		let now = Date.now();
		let msSinceLastTick = now % delay;

		this._timers[ channel ] = setTimeout(()=>{
			this.emit({str: 'Tick.', channel: channel, realtime: true});
			this._timers[ channel ] = setInterval(()=>{
				this.emit({str: 'Tick.', channel: channel, realtime: true});
			}, delay);
		}, delay - msSinceLastTick);

		return this;
	}

	unlisten(channel) {
		channel = Number(channel);
		if (channel <= 0) return;

		delete this._timers[ channel ];

		return this;
	}

	listChannels() {
		return ['day', 'hour', 'quarter', 'minute'];
	}

	getLastUpdateTimestamp() {
		// The clock runs updates real-time, so
		return Date.now();
	}
}


// signUp returns a Promise for a set of options. If the promise resolves, these options
// will be saved into the user preferences, then a new instance of this platform's `Account`
// will be instantiated.
export function signUp() {

	console.log('Performing sign-up procedure for a new "clock" account');

	// There is nothing asynchronous here, so...

	/// TODO: Extend this with a prompt for the default channel to listen to, or a prompt for
	/// the badge.

	return inquirer.prompt([{
		type: 'list',
		name: 'confirm',
		message: 'You are going to add a "clock" fake platform. Is this right?',
		choices: ['yes', 'no']
	}]).then((answer)=>{
		if (answer.confirm === 'yes') {
			return {
				// The clock platform does not need any parameters to initialize
				// an account. In a normal case, this response would include
				// auth tokens and so on.
				foo: 'bar',
				qux: { something: 'quux' }
			};
		} else {
			throw new Error('Not creating the new clock account then.');
		}
	});

// 	return Promise.resolve({
// 		// No options needed to initialize this platform, really.
// 		// Normally this would include auth tokens and such.
// 	});

}

export const defaultBadge = 'cl'; // or '⏰'
export const name = 'clock';
export { ClockAccount as Account };

