
// import Clock from './platforms/clock';

import * as platforms from './platforms/index'
import * as circularBuffer from './circularBuffer'
import stringify from './stringify'

import Vorpal from 'vorpal';
import inquirer from 'inquirer';
import Configstore from 'configstore';
import chalk from 'chalk256';
import Flatten from 'flatten-obj';
import TinyQueue from 'tinyqueue';
const flatten = Flatten();
const vorpal = Vorpal();

import deshortifyFilter from './filters/deshortify';

vorpal.history('soclial');
vorpal.log(chalk.purple('>>> Initializing soCLIal...'));

// create a Configstore instance with an unique ID e.g.
// Package name and optionally some default values
const conf = new Configstore('soclial', {
	global: {
		colours: {
			timestamp: 'limegreen',
			badge: 'pink',
			system: 'purple',
			updates: 'gray',
			media: 'cyan'
		}
	},
	accounts: {
	}
});

vorpal.log(chalk.purple('>>> User configuration file read.'));

const startTime = new Date();
const timezoneOffset = startTime.getTimezoneOffset();

// Define a tiniqueue for the items. This will allow fetching the Promise for
// the oldest update. Also the accounts can block the queue with synthetic
// promises that resolve to a falsy value.
// The queue will store entries in a { timestamp: 1234, itemPromise: Promise(...) } format.
function compareQueueItems(a, b) {
	return a.timestamp - b.timestamp;
}
let itemQueue = new TinyQueue([], compareQueueItems);



// This handles new item Promises from the accounts. Put them in the queue,
// wait for the oldest to resolve.
function handle(timestamp, itemPromise) {
// 	let { timestamp, channel } = item;

// 	console.log('handling: ', timestamp, itemPromise);
// 	console.log('handling: ', timestamp);

	// Pass the item through all known filters
	itemPromise = itemPromise.then((item)=>item && deshortifyFilter(item));

	itemQueue.push({ timestamp: timestamp, itemPromise: itemPromise });

	// Is this now the only item in the queue? Let's trigger a queue fetch in a short
	// while.
	if (itemQueue.length === 1) {
// // 		vorpal.log(chalk.purple('>>> Nothing in the queue, let\'s wait a bit to pop the oldest item.'));

		setTimeout(popOldest, 1000);
	}

}

// Pops the oldest in the queue, to stringify and display it.
// If there are no items left, do nothing. A new handle() is supposed to trigger
// this again.
function popOldest() {
	if (itemQueue.length) {
		/// TODO: Check if the oldest item is too close to the present, and
		/// delay it by a couple a seconds if so. At least delay things by the
		/// duration of an API roundtrip, to prevent out-of-order messages
		/// when streaming stuff.
// // 		vorpal.log(chalk.purple('>>> Popping item at ', itemQueue.peek().timestamp));

		itemQueue.pop().itemPromise.then(onItemReady);
	}
}


// stringifies and displays the item.
function onItemReady(item) {

	if (!item) {
		// If there is no item, this means that a synthetic entry was added
		// to the queue - most probably to block the queue temporarily.

		popOldest();	// Go to the next one.
		return;
	}

	if (item.timestamp === undefined) {
		item.timestamp = Date.now();
	}
	item.timestamp -= timezoneOffset * 60000;

	let id = circularBuffer.push(item);
	item.soclialId = id;

// 	console.log('onItemReady, item accountNumber is ', item.accountNumber);
// 	console.log('onItemReady, all accounts config is ', conf.all.accounts);
// 	console.log('onItemReady, account config is ', conf.all.accounts[item.accountNumber]);
	let badge = conf.all.accounts[Number(item.accountNumber)].badge;	// What's the badge for this platform/account??

	// https://stackoverflow.com/questions/10645994/node-js-how-to-format-a-date-string-in-utc
	let timestampString =
		new Date(item.timestamp)
			.toISOString()
			.replace(/T/, ' ')      // replace T with a space
			.replace(/\..+/, '');   // delete the dot and everything after


	const timestampColour = conf.get('global.colours.timestamp');
	const badgeColour     = conf.get('global.colours.badge');

	let header = '[' + chalk.foreground(timestampColour)(timestampString) + '] ' +
		id + '> ' + chalk.foreground(badgeColour)(badge + '/' + item.channel);

	let formatted = stringify(conf, item);

	// Printing out should NOT be synchronous with the handling.
	vorpal.log(header + formatted);

	popOldest();	// Go to the next one.
}





// Add a mock 'clock' account to the config
// conf.set('accounts.clock', {
// 	test: { // alias for the account
// 		channels: {
// 			minute: { /* channel-specific config */ },
// 			'10':   { /* channel-specific config */ },
// 		},
//
// 		/* other account-specific config */
// 		badge: 'clock'
// 	}
// });


// Iterate through the platforms/accounts/channels in the config, and go enabling them.


vorpal.log(chalk.purple('>>> Initializing social accounts...'));


// These define the state of the application in this global scope.
let activeAccounts = {};

function activateAccount(accId) {

// 	console.log(accId, conf.all.accounts[accId]);

	let accConf = conf.all.accounts[accId];
	let pl = accConf.platform;
	let badge = accConf.badge;

	vorpal.log(
		chalk.purple('>>> Activating account ') +
		chalk.cyan(accId) +
		chalk.purple(', represented by badge ') +
		chalk.cyan(badge) +
		chalk.purple(', platform ') +
		chalk.cyan(pl) +
		chalk.purple('...')
	);

	if (!((accConf.platform) in platforms)) {
		vorpal.log(chalk.red.bold('>>> Unknown platform: ') + chalk.cyan(pl));
		throw new Error('Unknown platform in a configured account. Please edit your configuration manually to remove the offending account.');
	}

	activeAccounts[accId] = new platforms[pl].Account(handle, vorpal, accId, accConf);

	for (let ch in accConf.channels) {
		vorpal.log(
			chalk.purple('>>> Listening to channel ') +
			chalk.cyan(ch) +
			chalk.purple('...')
		);
		activeAccounts[accId].listen(ch);
	}
}


for (let accId in conf.all.accounts) {
	activateAccount(accId);
}




// Command: config list
vorpal
	.command('config list', 'Lists the current configuration values.')
	.action(function(args, callback) {

		this.log( chalk.purple('>>> Listing current configuration:'));
		let all = flatten(conf.all);
		for (let i in all) {
			this.log( chalk.cyan(i) + ': ' + chalk.cyan(all[i]))
		}

		callback();
	});


// Command: config set
vorpal
	.command('config set <key> <value>', 'Sets a configuration value.')
	.action(function(args, callback) {
		if (args.key === undefined || args.value === undefined ) {
			this.log( chalk.red('>>> Specify configuration key and value to be set.'));
		} else {
			this.log(
				chalk.purple('>>> Setting configuration value ') +
				chalk.cyan(args.key) +
				chalk.purple(': ') +
				chalk.cyan(conf.get(args.key)) +
				chalk.purple(' → ') +
				chalk.cyan(args.value)
			);
			conf.set(args.key, args.value);
		}

		callback();
	});


// Command: config delete
vorpal
	.command('config delete <key>', 'Deletes a configuration value.')
	.action(function(args, callback) {

		if (args.key === undefined ) {
			this.log( chalk.red('>>> Specify configuration key to be deleted.'));
		} else {

			this.log(
				chalk.purple('>>> Deleting configuration value ') +
				chalk.cyan(args.key)
			);
			conf.delete(args.key);
		}

		callback();
	});






// Command: accounts list
vorpal
	.command('accounts list', 'Lists the accounts registered.')
	.action(function(args, callback) {
		this.log( chalk.purple('>>> Listing registered accounts:'));

		for (let accId in conf.all.accounts) {
			let accConf = conf.all.accounts[accId];
			let pl = accConf.platform;
			let badge = accConf.badge;

			vorpal.log(
				chalk.cyan(accId) +
				chalk.purple(', represented by badge ') +
				chalk.cyan(badge) +
				chalk.purple(', platform ') +
				chalk.cyan(pl)
			);
		}

		callback();
	});



// Command: accounts delete
vorpal
	.command('accounts delete <id>', 'Removes an account given its soCLIal numerical ID.')
	.action(function(args, callback) {

		if (args.id === undefined) {
			this.log( chalk.red('>>> Specify platform and alias of account to be deleted.'));
		} else {
			let flat = flatten(conf.all.accounts[args.id]);
			for (let i in flat) {
				conf.delete('accounts.' + args.id + '.' + i);
			}

			/// TODO: Unlisten to all channels
			/// TODO: Destroy instance

			this.log( chalk.red.bold('>>> Account deleted. Restart soCLIal to apply changes.'));
		}
		callback();
	});



// Command: accounts add
vorpal
	.command('accounts add', 'Adds a new account. You will be asked questions about it.')
	.action(function(args, callback) {

		vorpal.ui.cancel();

		inquirer.prompt([{
			type: 'list',
			name: 'platform',
			message: 'You are adding a new account to soCLIal. From which social platform?',
			choices: Object.keys(platforms)
		}]).then(answer=>{

// 			vorpal.ui.redraw();
			this.log('Asking the platform code to perform the sign-up procedure');

			return platforms[answer.platform].signUp().then((opts)=>{

				this.log('Platform code says we can create a new account with data: ', opts);

// 				vorpal.ui.cancel();

				return inquirer.prompt([{
					type: 'input',
					name: 'badge',
					message: 'What will be the badge of this account? The badge is a string of text that will show next to all messages from this account, and it\'s better to keep it short.',
					choices: Object.keys(platforms),
					default: platforms[answer.platform].badge
				}]).then(answer2=>{

					opts.badge = answer2.badge;
					opts.platform = answer.platform;

					// Find the lowest unused account ID
					let accId = 1;
					while( accId in conf.all.accounts ) {
						accId++;
					}

	// 					this.log( chalk.red.bold('>>> NOT IMPLEMENTED YET.'));
					this.log( 'Should store config for a new account: ', opts);
	// 					ui.redraw.done()

					let flat = flatten(opts);

					this.log(conf.all.accounts);
					this.log('flat account config: ', flat);

					for (let i in flat) {
						conf.set('accounts.' + accId + '.' + i , flat[i]);
						this.log('accounts.' + accId + '.' + i + ' = ' + flat[i]);
					}

					this.log( 'New account added to config. Enabling now.');

					activateAccount(accId);

					callback();

// 				}).catch((err)=>{
// 					this.log( chalk.red.bold('>>> Creation of new account cancelled (no badge).', err));
// 	// 					ui.redraw.done()
// 					callback();
				});
// 			}).catch((err)=>{
//
// 				this.log( chalk.red.bold('>>> Creation of new account cancelled (no data from sign-up procedure).', err));
// // 					ui.redraw.done()
// 				callback();
			});
		}).catch((err)=>{
			this.log( chalk.red.bold('>>> Creation of new account cancelled ', err));
// 			ui.redraw.done()
			callback();
		});

	});



// Command: item url
vorpal
	.command('url <id>', 'Displays the unequivocal URL for an item, given its soCLIal letter-letter-number ID.')
	.action(function(args, callback) {

		if (args.id === undefined) {
			this.log( chalk.red('>>> Specify item ID, e.g. "url ac7".'));
		} else {

			let item = circularBuffer.get(args.id);

			if (!item) {
				this.log( chalk.red('>>> Item ID doesn\'t exist.'));
			} else {
				this.log(
					chalk.purple('>>> URL for ') +
					chalk.cyan(args.id) +
					chalk.purple(' is ') +
					chalk.cyan(item.url)
				);
			}
		}
		callback();
	});


// Command: post something
vorpal
	.command('post [str...]')
	.description('Post something to all platforms at once.')
	.action((args, callback)=>{
		let str = args.str.join(' ');
		vorpal.log(chalk.green('Should post: ' + args.str.join(' ')));
// 		this.cancel();

		for (let i in activeAccounts) {
			activeAccounts[i].post(str);
		}

		callback();
	});


// Command: reply to something
vorpal
	.command('reply <id> [str...]')
	.description('Post a reply to an item.')
	.action((args, callback)=>{

		if (args.id === undefined) {
			this.log( chalk.red('>>> Specify item ID, e.g. "url ac7".'));
		} else {
			let item = circularBuffer.get(args.id);
			let str = args.str.join(' ');


			vorpal.log(chalk.green('Should reply with: ' + args.str.join(' ')));
	// 		this.cancel();

			activeAccounts[Number(item.accountNumber)].reply(item, str);
		}
		callback();
	});

// Command: echo something
vorpal
	.command('echo <id>')
	.description('Echo (RT/boost/reshare) an item, in all platforms at once.')
	.action((args, callback)=>{

		if (args.id === undefined) {
			this.log( chalk.red('>>> Specify item ID, e.g. "url ac7".'));
		} else {
			let item = circularBuffer.get(args.id);

			let fallbackStr = activeAccounts[Number(item.accountNumber)].echoify(item);
			vorpal.log(chalk.green('String for fallback echo is: ' + fallbackStr));

			for (let i in activeAccounts) {
				activeAccounts[i].echo(item, fallbackStr);
			}
		}
		callback();
	});


vorpal.log( chalk.white.bold('>>> Welcome to soCLIal !!!') );
vorpal.log( chalk.white.bold('>>> Type "help" to see available commands.') );



// Init main loop.
vorpal
	.delimiter('soCLIal> ')
	.show();





